/*

findere.c - Find Extended Regular Expression

*/

/*...sincludes:0:*/
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <memory.h>
#include <sys/types.h>
#include <sys/stat.h>
#if defined(O16) || defined(OS2) || defined(WIN32)
#include "io.h" /* for isatty */
#endif

#include "ere.h"
#ifdef USE_DIRT
#include "dirt.h"
#endif
/*...e*/

/*...sFE_BOOLEAN:0:*/
#ifndef FE_BOOLEAN_DEFINED
#define	FE_BOOLEAN_DEFINED
typedef	int FE_BOOLEAN;
#define	FE_TRUE  1
#define	FE_FALSE 0
#endif
/*...e*/

static char progname[] = "findere";

/*...sfatal:0:*/
static void fatal(const char *fmt, ...)
	{
	va_list	vars;
	char s[256+1];
	va_start(vars, fmt);
	vsprintf(s, fmt, vars);
	va_end(vars);
	fprintf(stderr, "%s: %s\n", progname, s);
	exit(1);
	}
/*...e*/
/*...sinfo:0:*/
static void info(const char *fmt, ...)
	{
	va_list	vars;
	char s[256+1];
	va_start(vars, fmt);
	vsprintf(s, fmt, vars);
	va_end(vars);
	fprintf(stderr, "%s: %s\n", progname, s);
	}
/*...e*/
/*...susage:0:*/
static void usage(void)
	{
#ifdef USE_DIRT
	fprintf(stderr, "usage: %s [-n ere|-N glob] [-x ere|-X glob] [-y ere|-Y glob] [-r]\n", progname);
#else
	fprintf(stderr, "usage: %s [-n ere|-N glob] [-x ere|-X glob]\n", progname);
#endif
	fprintf(stderr, "               [-j|-J] [-i] [-m] [-s] [-v] [-c]\n");
	fprintf(stderr, "               [f|-F] [-l] [-S] [-a|-A] [-q] [--] ere {fn}\n");
	fprintf(stderr, "flags: -n ere   search files whose base filenames match ere\n");
	fprintf(stderr, "       -N glob  search files whose base filenames match UNIX style glob\n");
	fprintf(stderr, "       -x ere   exclude files whose basenames match ere\n");
	fprintf(stderr, "       -X glob  exclude files whose basenames match UNIX style glob\n");
#ifdef USE_DIRT
	fprintf(stderr, "       -y ere   exclude directories whose basenames match ere\n");
	fprintf(stderr, "       -Y glob  exclude directories whose basenames match UNIX style glob\n");
	fprintf(stderr, "       -r       recurse into directories\n");
#endif
	fprintf(stderr, "       -j -J    case insensitive pathname match, or sensitive\n");
#ifdef UNIX
	fprintf(stderr, "                (default: sensitive on UNIX)\n");
#else
	fprintf(stderr, "                (default: insensitive on non-UNIX)\n");
#endif
	fprintf(stderr, "       -i       case insensitive search\n");
	fprintf(stderr, "       -m       don't look for multiple matches on each line\n");
	fprintf(stderr, "       -s       look for shortest match (default is longest)\n");
	fprintf(stderr, "       -v       look for lines which don't match the ere\n");
	fprintf(stderr, "       -c       display count of matches, not matches themselves\n");
	fprintf(stderr, "       -f -F    display filenames with matches or counts, or don't\n");
#ifdef USE_DIRT
	fprintf(stderr, "                (default: show if more than one filename given, or -r)\n");
#else
	fprintf(stderr, "                (default: show if more than one filename given)\n");
#endif
	fprintf(stderr, "       -l       display line number before each match\n");
	fprintf(stderr, "       -S       show submatches (implies no -c, no -v and -m)\n");
	fprintf(stderr, "       -a -A    highlight output in colour, or don't\n");
#if defined(O16) || defined(OS2)
	fprintf(stderr, "                (default: use ANSI, if tty)\n");
#elif defined(WIN32)
	fprintf(stderr, "                (default: use ConsoleAPI, if tty)\n");
#elif defined(UNIX)
	fprintf(stderr, "                (default: use ANSI, if tty and recognised terminal type)\n");
#else
	fprintf(stderr, "                (default: don't use ANSI)\n");
#endif
	fprintf(stderr, "       -q       don't show some error messages\n");
	fprintf(stderr, "       ere      extended regular expression\n");
#ifdef xxDOS
	fprintf(stderr, "                (on DOS systems, for | use !)\n");
#endif
	fprintf(stderr, "       fn       filename(s), no extension assumed (- for stdin)\n");
	exit(1);
	}
/*...e*/
/*...scolour_\42\:0:*/
/* Here we abstract away the platform specific details of colour hilighting. */

static FE_BOOLEAN colour = FE_FALSE;

#ifdef WIN32
#define	WIN32_LEAN_AND_MEAN
#include <windows.h>
#undef WIN32_LEAN_AND_MEAN
static HANDLE hConOut;
static CONSOLE_SCREEN_BUFFER_INFO csbiSaved;
static CHAR_INFO chi;
#endif

static FE_BOOLEAN colour_autodetect(void)
	{
#if defined(WIN32)
	if ( _isatty(_fileno(stdout)) )
		return FE_TRUE;
#elif defined(O16) || defined(OS2)
	if ( isatty(fileno(stdout)) )
		return FE_TRUE;
#elif defined(UNIX)
	if ( isatty(fileno(stdout)) )
		{
		const char *term;
		if ( (term = getenv("TERM")) != NULL )
			if ( !memcmp(term, "vt", 2)    ||
			     !memcmp(term, "xterm", 5) ||
			     !strcmp(term, "linux")    ||
			     !strcmp(term, "aixterm")  ||
			     !strcmp(term, "hft")      )
				return FE_TRUE;
		}
#endif
	return FE_FALSE;
	}

static void colour_init(void)
	{
	if ( colour )
#ifdef WIN32
		{
		hConOut = GetStdHandle(STD_OUTPUT_HANDLE);
		GetConsoleScreenBufferInfo(hConOut, &csbiSaved);
		SetConsoleTextAttribute(hConOut,
			FOREGROUND_RED  |
			FOREGROUND_GREEN|
			FOREGROUND_BLUE );
		}
#else
		fputs("\033[37m", stdout); /* White */
#endif
	}

static void colour_normal(void)
	{
	if ( colour )
#ifdef WIN32
		SetConsoleTextAttribute(hConOut,
			FOREGROUND_RED  |
			FOREGROUND_GREEN|
			FOREGROUND_BLUE );
#else
		fputs("\033[37m", stdout); /* White */
#endif
	}

static void colour_file_prefix(void)
	{
	if ( colour )
#ifdef WIN32
		SetConsoleTextAttribute(hConOut,
			FOREGROUND_GREEN|
			FOREGROUND_BLUE );
#else
		fputs("\033[36m", stdout); /* Cyan */
#endif
	}

static void colour_num_prefix(void)
	{
	if ( colour )
#ifdef WIN32
		SetConsoleTextAttribute(hConOut,
			FOREGROUND_RED);
#else
		fputs("\033[31m", stdout); /* Red */
#endif
	}

static void colour_match(int n)
	{
	if ( colour )
#ifdef WIN32
		SetConsoleTextAttribute(hConOut,
			(WORD) ( ( (n&1) == 0 )
				? ( FOREGROUND_GREEN )
				: ( FOREGROUND_RED   |
				    FOREGROUND_GREEN ) ) );
#else
		fputs( ( (n&1) == 0 ) ? "\033[32m" : "\033[33m", stdout);
			/* Green or yellow */
#endif
	}

static void colour_term(void)
	{
#ifdef WIN32
	if ( colour )
		SetConsoleTextAttribute(hConOut, csbiSaved.wAttributes);
#endif
	}
/*...e*/
/*...smain:0:*/
#ifdef xxDOS
#define	MAX_FN 500
#define	MAX_GLOB 250
#else
#define	MAX_FN 2000
#define	MAX_GLOB 1000
#endif

#ifdef UNIX
#define	SEP_CHAR '/'
#define	SEP_STR  "/"
#else
#define	SEP_CHAR '\\'
#define	SEP_STR  "\\"
#endif

#ifdef USE_DIRT
static FE_BOOLEAN recurse     = FE_FALSE;
#endif
#ifdef UNIX
static FE_BOOLEAN insensitive_fn = FE_FALSE;
#else
static FE_BOOLEAN insensitive_fn = FE_TRUE;
#endif
static FE_BOOLEAN insensitive    = FE_FALSE;
static FE_BOOLEAN multiple       = FE_TRUE;
static FE_BOOLEAN shortest       = FE_FALSE;
static FE_BOOLEAN negate         = FE_FALSE;
static FE_BOOLEAN count_mode     = FE_FALSE;
static FE_BOOLEAN show_file;
static FE_BOOLEAN show_line      = FE_FALSE;
static FE_BOOLEAN show_subs      = FE_FALSE;
static FE_BOOLEAN quiet          = FE_FALSE;

#ifdef xxDOS
#define	MAX 500
#else
#define	MAX 2000
#endif

static ERE *ere;
static ERE *ere_pattern;
static ERE *ere_exclude;
static ERE *ere_exclude_dir;

/*...stolower_str:0:*/
static char *tolower_str(char *s)
	{
	char *p;
	for ( p = s; *p; p++ )
		*p = (char) tolower(*p);
	return s;
	}
/*...e*/
/*...sglob_to_ere:0:*/
/* Facilitate the support of UNIX style file wildcard "globs".
   We do this by trivially converting them to EREs. */

static FE_BOOLEAN glob_to_ere(const char *glob, char *ere)
	{
	while ( *glob )
		switch ( *glob )
			{
			case '?':
				*ere++ = '.';
				glob++;
				break;
			case '*':
				*ere++ = '.';
				*ere++ = '*';
				glob++;
				break;
			case '.':
			case '(':
			case ')':
			case '{':
			case '}':
			case '+':
			case '|':
			case '~':
			case '^':
			case '$':
				*ere++ = '\\';
				*ere++ = *glob;
				glob++;
				break;
			case '[':
				*ere++ = *glob++;
				while ( *glob != ']' )
					switch ( *glob )
						{
						case '\\':
							*ere++ = *glob++;
							if ( *glob == '\0' )
								return FE_FALSE;
							*ere++ = *glob++;
							break;
						case '\0':
							return FE_FALSE;
						default:
							*ere++ = *glob++;
							break;
						}
				*ere++ = *glob++;
				break;
			case '\\':
				*ere++ = *glob++;
				*ere++ = *glob++;
				break;
			default:
				*ere++ = *glob++;
				break;
			}
	*ere = '\0';
	return FE_TRUE;
	}
/*...e*/
/*...sfindere_stream:0:*/
/*...smatches:0:*/
static FE_BOOLEAN matches(const char *line, int start, int *from, int *to, ERE_MATCHINFO *mi)
	{
	int pos_match, len_match;
	if ( !ere_match_fwd(ere,
			    shortest ? EREMF_SHORTEST : 0,
			    line, start, &pos_match, &len_match, mi) ) 
		return FE_FALSE;
	*from = pos_match;
	*to   = pos_match + len_match;
	return FE_TRUE;
	}
/*...e*/
/*...sfile_prefix:0:*/
static void file_prefix(const char *filename)
	{
	if ( show_file )
		{
		fputs("\n--- ", stdout);
		colour_file_prefix();
		fputs(filename, stdout);
		colour_normal();
		fputs(" ---\n\n", stdout);
		}
	}
/*...e*/
/*...sline_prefix:0:*/
static void line_prefix(unsigned long line_no)
	{
	if ( show_line )
		{
		colour_num_prefix();
		printf("%7lu\t", line_no);
		colour_normal();
		}
	}
/*...e*/
/*...sshow:0:*/
static void show(const char *str, int n)
	{
	while ( n-- )
		putchar(*str++);
	}
/*...e*/

static void findere_stream(FILE *fp, const char *filename)
	{
	unsigned long line_no = 0L;
	char line[MAX+1];
	long count = 0L;
	ERE_MATCHINFO mi2;
	ERE_MATCHINFO *mi = ( show_subs ) ? &mi2 : NULL;

	while ( fgets(line, MAX + 1, fp) != NULL )
		{
		char lcase_line[MAX+1], *l, *l2;
		int len, from, to;

		line_no++;
		len = strlen(line);
		if ( line[len-1] == '\n' )
			line[--len] = '\0';
		if ( insensitive )
			l = tolower_str(strcpy(lcase_line, line));
		else
			l = line;
		l2 = line;
		if ( negate )
			/* We're interested in those which don't match */
			{
			if ( !matches(l, 0, &from, &to, mi) )
				{
				++count;
				if ( !count_mode )
					{
					if ( count == 1 )
						file_prefix(filename);
					line_prefix(line_no);
					fputs(l2, stdout);
					putchar('\n');
					}
				}
			}
		else
			{
			if ( matches(l, 0, &from, &to, mi) )
				{
				++count;
				if ( !count_mode )
					{
					if ( count == 1 )
						file_prefix(filename);
					line_prefix(line_no);
					show(l2, from);
					colour_match(0);
					show(l2+from, to-from);
					colour_normal();
					}
				if ( multiple )
					{
					int count_prev = 0;
					int start = to;
					while ( matches(l, start+(from==to), &from, &to, mi) )
						{
						++count;
						if ( !count_mode )
							{
							show(l2+start, from-start);
							colour_match(++count_prev);
							show(l2+from, to-from);
							colour_normal();
							}
						start = to;
						}
					}
				if ( !count_mode )
					{
					fputs(l2+to, stdout);
					putchar('\n');
					}
				if ( mi != NULL )
					{
					int i;
					for ( i = 0; i < mi->n_spans; i++ )
						{
						printf("\t\\%u=", i+1);
						colour_match(1);
						show(l2+mi->spans[i].pos,
						     mi->spans[i].len);
						colour_normal();
						fputs("\n", stdout);
						}
					}
				}
			}
		}

	if ( count_mode )
		{
		printf("%7lu", count);
		if ( show_file )
			printf(" %s", filename);
		fputs("\n", stdout);
		}
	}
/*...e*/
/*...sfindere:0:*/
static void findere(const char *filename)
	{
	struct stat buf;

#ifdef xxDOS
	if ( stat((char *) filename, &buf) != -1 &&
#else
	if ( stat(         filename, &buf) != -1 &&
#endif
	     (buf.st_mode & S_IFDIR) != 0 )
		/* Its a directory */
		{
#ifdef USE_DIRT
		if ( ere_exclude_dir != NULL )
			/* Had better validate the directory before searching */
			{
			char *p, fn[MAX_FN+1];
			int len;
			if ( (p = strrchr(filename, SEP_CHAR)) != NULL )
				strcpy(fn, p+1);
			else
				strcpy(fn, filename);
			if ( insensitive_fn )
				tolower_str(fn);
			len = ere_match(ere_exclude_dir, 0, fn, 0, NULL);
			if ( len == (int) strlen(fn) )
				return; /* excluded */
			}
		if ( recurse )
			{
			DIRT *dirt;
			int rc;
			if ( (dirt = dirt_open(filename, &rc)) == NULL )
				{
				if ( !quiet )
					info("can't open %s: %s", filename, dirt_error(rc));
				}
			else
				{
				const char *fn;
				while ( (fn = dirt_next(dirt)) != NULL )
					if ( strcmp(fn, ".") && strcmp(fn, "..") )
						{
						if ( strlen(filename) +
						     strlen(SEP_STR) +
						     strlen(fn) > MAX_FN )
							{
							if ( !quiet )
								info("pathname too long");
							}
						else
							{
							char nested[MAX_FN+1];
							int len = strlen(filename);
							strcpy(nested, filename);
							if ( nested[len-1] != '/' && nested[len-1] != '\\' )
								strcpy(nested+len, SEP_STR);
							strcat(nested, fn);
							findere(nested);
							}
						}
				dirt_close(dirt);
				}
			}			
		/* else
			silently ignore it */
#endif
		}
	else
		{
		FILE *fp;
		if ( ere_exclude != NULL )
			/* Had better validate the filename before searching */
			{
			char *p, fn[MAX_FN+1];
			int len;
			if ( (p = strrchr(filename, SEP_CHAR)) != NULL )
				strcpy(fn, p+1);
			else
				strcpy(fn, filename);
			if ( insensitive_fn )
				tolower_str(fn);
			len = ere_match(ere_exclude, 0, fn, 0, NULL);
			if ( len == (int) strlen(fn) )
				return; /* excluded */
			}
		if ( ere_pattern != NULL )
			/* Had better validate the filename before searching */
			{
			char *p, fn[MAX_FN+1];
			int len;
			if ( (p = strrchr(filename, SEP_CHAR)) != NULL )
				strcpy(fn, p+1);
			else
				strcpy(fn, filename);
			if ( insensitive_fn )
				tolower_str(fn);
			len = ere_match(ere_pattern, 0, fn, 0, NULL);
			if ( len != (int) strlen(fn) )
				return;
			}
		if ( (fp = fopen(filename, "r")) == NULL )
			{
			if ( !quiet )
				info("can't open %s for reading", filename);
			}
		else
			{
			findere_stream(fp, filename);
			fclose(fp);
			}
		}
	}
/*...e*/

int main(int argc, char *argv [])
	{
	char *pattern;
	char *pattern_fn = NULL;
	char pattern_ere[MAX_GLOB+1];
	char *exclude_fn = NULL;
	char exclude_ere[MAX_GLOB+1];
	char *exclude_fn_dir = NULL;
	char exclude_ere_dir[MAX_GLOB+1];
	int i, rc;
	FE_BOOLEAN df = FE_TRUE, dc = FE_TRUE;

	if ( argc == 1 )
		usage();

	for ( i = 1; i < argc; i++ )
		{
		if ( argv[i][0] != '-' )
			break;
		else if ( argv[i][1] == '-' )
			{ ++i; break; }
		switch ( argv[i][1] )
			{
			case 'n':
				if ( ++i == argc )
					usage();
				pattern_fn = argv[i];
				break;
			case 'N':
				if ( ++i == argc )
					usage();
				if ( !glob_to_ere(argv[i], pattern_ere) )
					fatal("badly formed UNIX pattern glob");
				pattern_fn = pattern_ere;
				break;
			case 'x':
				if ( ++i == argc )
					usage();
				exclude_fn = argv[i];
				break;
			case 'X':
				if ( ++i == argc )
					usage();
				if ( !glob_to_ere(argv[i], exclude_ere) )
					fatal("badly formed UNIX exclude glob");
				exclude_fn = exclude_ere;
				break;
#ifdef USE_DIRT
			case 'y':
				if ( ++i == argc )
					usage();
				exclude_fn_dir = argv[i];
				break;
			case 'Y':
				if ( ++i == argc )
					usage();
				if ( !glob_to_ere(argv[i], exclude_ere_dir) )
					fatal("badly formed UNIX exclude glob");
				exclude_fn_dir = exclude_ere_dir;
				break;
			case 'r': recurse = FE_TRUE;			break;
#endif
			case 'j': insensitive_fn = FE_TRUE;		break;
			case 'J': insensitive_fn = FE_FALSE;		break;
			case 'i': insensitive = FE_TRUE;		break;
			case 'm': multiple = FE_FALSE;			break;
			case 'v': negate = FE_TRUE;			break;
			case 's': shortest = FE_TRUE;			break;
			case 'c': count_mode = FE_TRUE;			break;
			case 'f': show_file = FE_TRUE; df = FE_FALSE;	break;
			case 'F': show_file = FE_FALSE; df = FE_FALSE;	break;
			case 'l': show_line = FE_TRUE;			break;
			case 'a': colour = FE_TRUE; dc = FE_FALSE;	break;
			case 'A': colour = FE_FALSE; dc = FE_FALSE;	break;
			case 'S': show_subs = FE_TRUE;			break;
			case 'q': quiet = FE_TRUE;			break;
			default:  usage();				break;
			}
		}

	if ( show_subs )
		{
		count_mode = FE_FALSE;
		negate      = FE_FALSE;
		multiple   = FE_FALSE;
		}

	if ( i == argc )
		usage();
	pattern = argv[i++];

	/* Compile the ERE to match */
	if ( (ere = ere_compile(pattern,
				insensitive ? ERECF_TOLOWER : 0,
				&rc)) == NULL )
		fatal("bad extended regular expression (%s)", ere_error(rc));

	/* Compile the filename matching ERE */
	if ( pattern_fn != NULL )
		{
		if ( (ere_pattern = ere_compile(pattern_fn,
					insensitive_fn ? ERECF_TOLOWER : 0,
					&rc)) == NULL )
			fatal("bad pathname extended regular expression (%s)", ere_error(rc));
		}
	else
		ere_pattern = NULL;

	/* Compile the filename excluding ERE */
	if ( exclude_fn != NULL )
		{
		if ( (ere_exclude = ere_compile(exclude_fn,
					insensitive_fn ? ERECF_TOLOWER : 0,
					&rc)) == NULL )
			fatal("bad exclude filename extended regular expression (%s)", ere_error(rc));
		}
	else
		ere_exclude = NULL;

	/* Compile the directory excluding ERE */
	if ( exclude_fn_dir != NULL )
		{
		if ( (ere_exclude_dir = ere_compile(exclude_fn_dir,
					insensitive_fn ? ERECF_TOLOWER : 0,
					&rc)) == NULL )
			fatal("bad exclude directory extended regular expression (%s)", ere_error(rc));
		}
	else
		ere_exclude_dir = NULL;

#ifdef xxDOS
/*...scannot enter \124\ on dos command line:8:*/
{
char *p;
for ( p = pattern; *p; p++ )
	if ( *p == '!' )
		*p = '|';
}
/*...e*/
#endif

	if ( dc )
		colour = colour_autodetect();

	if ( df )
		{
		show_file = ( i+1 < argc );
#ifdef USE_DIRT
		show_file |= recurse;
#endif
		}

	colour_init();

	for ( ; i < argc; i++ )
		if ( !strcmp(argv[i], "-") )
			findere_stream(stdin, "stdin");
		else
			findere(argv[i]);

	if ( ere_pattern != NULL )
		ere_discard(ere_pattern);
	if ( ere_exclude != NULL )
		ere_discard(ere_exclude);
	if ( ere_exclude_dir != NULL )
		ere_discard(ere_exclude_dir);
	ere_discard(ere);

	colour_term();
	return 0;
	}
/*...e*/
